/************************************************************************
 *   IRC - Internet Relay Chat, src/cmodes.c
 *   This file is copyright (C) 2001 Andrew Suffield
 *                                    <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "config.h"
#include "channel.h"
#include "client.h"
#include "cmodes.h"
#include "common.h"
#include "irc_string.h"
#include "ircd.h"
#include "list.h"
#include "memory.h"
#include "numeric.h"
#include "s_log.h"
#include "send.h"
#include "struct.h"

#include <stdlib.h>
#include <assert.h>

struct channel_mode cmodes[] = {
  /* Final digit is 1 if the umode should be sorted to the end of the list
   *  when building strings to send the the client
   */
  {CUFLAG_CHANOP,	'o', CMTYPE_USERFLAG,	UMODE_INVALID, 0},
  {CUFLAG_VOICE,	'v', CMTYPE_USERFLAG,	UMODE_INVALID, 0},
  {CMODE_SECRET,	's', CMTYPE_CHANFLAG,	UMODE_INVALID, 0},
  {CMODE_MODERATED,	'm', CMTYPE_CHANFLAG,	UMODE_INVALID, 0},
  {CMODE_TOPICLIMIT,	't', CMTYPE_CHANFLAG,	UMODE_INVALID, 0},
  {CMODE_INVITEONLY,	'i', CMTYPE_CHANFLAG,	UMODE_INVALID, 0},
  {CMODE_NOPRIVMSGS,	'n', CMTYPE_CHANFLAG,	UMODE_INVALID, 0},
  {RPL_BANLIST,		'b', CMTYPE_MASK,	UMODE_INVALID, 0},
  {RPL_EXCEPTLIST,	'e', CMTYPE_MASK,	UMODE_INVALID, 1},
  {RPL_INVITELIST,	'I', CMTYPE_MASK,	UMODE_INVALID, 1},
  {-1,			'l', CMTYPE_INTVAL,	UMODE_INVALID, 0},
  {-1,			'f', CMTYPE_STRVAL,	UMODE_INVALID, 1},
  {CMODE_NOCOLOUR,	'c', CMTYPE_CHANFLAG,	UMODE_INVALID, 1},
  {CMODE_PERM,		'P', CMTYPE_CHANFLAG,	UMODE_EXPERIMENTAL, 1},
  {CMODE_ALLOWFORWARD,	'F', CMTYPE_CHANFLAG,	UMODE_EXPERIMENTAL, 1},
  {CMODE_JUPED,		'j', CMTYPE_CHANFLAG,	UMODE_GOD, 1},
  {-1,			'q', CMTYPE_MASK,	UMODE_INVALID, 1},
  {0,			'\0', 0, 0, 0} /* Last entry delimited by \0 character */
};

int cmode_index[256];		/* letter -> index in channel_modes */
static int mask_reply[CMODE_MASK_COUNT], mask_end_reply[CMODE_MASK_COUNT];
static char mask_index_char[CMODE_MASK_COUNT];
static char modebuf[MODEBUFLEN], modebuf2[MODEBUFLEN];
static char parabuf[MODEBUFLEN], parabuf2[MODEBUFLEN];

void
init_cmode_tables(void)
{
  unsigned int i;
  int intval_count = 0, strval_count = 0, mask_count = 0;
  for (i = 0; i < 256; i++)
    cmode_index[i] = -1;
  for (i = 0; cmodes[i].letter; i++)
    {
      unsigned char c = cmodes[i].letter;
      cmode_index[c] = i;
      switch (cmodes[i].type)
	{
	case CMTYPE_INTVAL:
	  cmodes[i].value = intval_count++;
	  break;
	case CMTYPE_STRVAL:
	  cmodes[i].value = strval_count++;
	  break;
	case CMTYPE_MASK:
	  mask_reply[mask_count] = cmodes[i].value;
	  mask_end_reply[mask_count] = cmodes[i].value + 1;
	  mask_index_char[mask_count] = c;
	  cmodes[i].value = mask_count++;
	  break;
	}
    }
  if (intval_count > CMODE_INTVAL_COUNT)
    {
      log(L_CRIT,
	  "%d intvals in cmodes[], but only %d allocated in cmodes.h",
	  intval_count, CMODE_INTVAL_COUNT);
      exit(-1);
    }
  if (strval_count > CMODE_STRVAL_COUNT)
    {
      log(L_CRIT,
	  "%d strvals in cmodes[], but only %d allocated in cmodes.h",
	  strval_count, CMODE_STRVAL_COUNT);
      exit(-1);
    }
  if (mask_count > CMODE_MASK_COUNT)
    {
      log(L_CRIT, "%d masks in cmodes[], but only %d allocated in cmodes.h",
	  mask_count, CMODE_MASK_COUNT);
      exit(-1);
    }
}

/*
 * like the name says...  take out the redundant signs in a modechange list
 */
void
collapse_signs(char *s)
{
  char plus = '\0', *t = s, c;
  while ((c = *s++))
    {
      if (c != plus)
	*t++ = c;
      if (c == '+' || c == '-')
	plus = c;
    }
  *t = '\0';
}

int
parse_channel_mask(const char *c, struct channel_mask *mask, int *parc,
		   char ***parv)
{
  /* OK, we decompose ban masks from the _right_, according to this
   * structure:
   * nick!user@host!#channel%timeout;"reason"
   * where all elements are optional.
   *
   * reason may contain spaces
   *
   * Missing elements have the following effects:
   *
   * nick:    Match any nick
   * user:    Match any user ) In the case where neither user nor host is given, 
   * host:    Match any host ) match the _registered_ nick unless it contains wilds
   * channel: No forwarding to another channel. Not all masks support this.
   * timeout: Mask remains until it is manually removed
   * reason:  No effect
   */

  char *pos;
  char reasonbuf[BUFSIZE];	/* For simplicity, with the multi-word stuff. Nothing can overflow this */

  /* OK, anything following a ; is a reason, if we have one */
  if ((pos = strrchr(c, ';')))
    {
      *pos++ = '\0';
      if (*pos == '"')
	{
	  strcpy(reasonbuf, pos + 1);
	  /* This is a little ineffecient, because it scans the whole string
	   *  repeatedly, but it's good enough. (doesn't get called often)
	   *  -- asuffield
	   */
	  while (!(pos = strchr(reasonbuf, '"')) && *parc)
	    {
	      strcat(reasonbuf, " ");
	      strcat(reasonbuf, (*parv)[0]);
	      (*parv)++;
	      (*parc)--;
	    }
	  if (pos)
	    *pos = '\0';

	  /* Truncate regardless */
	  strncpy_irc(mask->reason, reasonbuf, REASONLEN);
	}
      else
	strncpy_irc(mask->reason, pos, REASONLEN);
    }
  else
    mask->reason[0] = '\0';

  /* That's the hard part over */
  if ((pos = strrchr(c, '%')))
    {
      *pos++ = '\0';
      /* If the number is +n, then it's n seconds from now */
      if (*pos == '+')
	mask->timeout = CurrentTime + atoi(pos + 1);
      /* Otherwise it's n seconds from the epoch */
      else
	mask->timeout = atoi(pos);
    }
  else
    mask->timeout = 0;

  if ((pos = strrchr(c, '!')) && (*(pos + 1) == '#'))
    {
      *pos++ = '\0';
      strncpy_irc(mask->channel, pos, CHANNELLEN);
    }
  else
    mask->channel[0] = '\0';

  if ((pos = strrchr(c, '@')))
    {
      *pos++ = '\0';
      strncpy_irc(mask->host, pos, HOSTLEN);
    }
  else
    mask->host[0] = '\0';

  if ((pos = strrchr(c, '!')))
    {
      *pos++ = '\0';
      strncpy_irc(mask->user, pos, USERLEN);
    }
  else
    mask->user[0] = '\0';

  strncpy_irc(mask->nick, c, NICKLEN);

  /* OK, some of those fields are "" and they might need to be "*" */
  if (!*mask->nick)
    {
      mask->nick[0] = '*';
      mask->nick[1] = '\0';
    }

  if (!*mask->user && *mask->host)
    {
      mask->user[0] = '*';
      mask->user[1] = '\0';
    }

  if (!*mask->host && *mask->user)
    {
      mask->host[0] = '*';
      mask->host[1] = '\0';
    }

  return 0;
}

char *
build_channel_mask(const struct channel_mask *mask)
{
  static char buf[BUFSIZE];

  /* Gah. Is there not an easier way to do this? */
  strncpy_irc(buf, mask->nick, NICKLEN);
  if (*mask->user)
    {
      strcat(buf, "!");
      strcat(buf, mask->user);
    }
  if (*mask->host)
    {
      strcat(buf, "@");
      strcat(buf, mask->host);
    }
  if (*mask->channel)
    {
      strcat(buf, "!");
      strcat(buf, mask->channel);
    }
  if (mask->timeout)
    {
      strcat(buf, "%");
      ircsprintf(buf + strlen(buf), "%.1lu", mask->timeout);
    }
  if (*mask->reason)
    {
      strcat(buf, ";\"");
      strcat(buf, mask->reason);
      strcat(buf, "\"");
    }

  return buf;
}

struct channel_mask *
build_mask(const struct Client *cptr)
{
  static struct channel_mask mask;

  if (HasUmode(cptr, UMODE_IDENTIFIED))
    {
      strncpy_irc(mask.nick, cptr->name, NICKLEN);
      mask.user[0] = '\0';
      mask.host[0] = '\0';
    }
  else
    {
      strncpy_irc(mask.nick, cptr->name, NICKLEN);
      strncpy_irc(mask.user, cptr->username, USERLEN);
      strncpy_irc(mask.host, cptr->host, HOSTLEN);
    }

  return &mask;
}

struct SLink *
find_channel_mask(const struct channel_mask *search_mask,
		  struct Channel *chptr, u_int32_t type, int exact,
		  int expire)
{
  struct SLink *cmlink, **prev;

  /* We are going to search for mask, deleting dead entries along the way */

  /* This complex beastie is a look-behind-one setup. prev always points to the last
   * 'next' field, so that it can be updated if the _current_ struct (cmlink) is deleted
   *
   * The extra trickery here is that it doesn't crash in the case where cmlink is NULL,
   * but instead exits smoothly. This means we don't have to break after deleting an
   * item.
   */
  for (prev = &(chptr->mode_mask_list[type]); (cmlink = *prev);
       prev = (cmlink ? &(cmlink->next) : &cmlink))
    {
      struct channel_mask *mask = cmlink->value.cmptr;

      if (!cmlink->value.cmptr)
	{
	  /* This is a dead link that del_channel_mask() left behind,
	   * clean it up
	   */
	  *prev = cmlink->next;
	  free_link(cmlink);
	  cmlink = *prev;
	  continue;
	}

      if (expire && cmlink->value.cmptr->timeout
	  && cmlink->value.cmptr->timeout < CurrentTime)
	{
	  /* This one has expired */
	  sendto_channel_butserv(chptr, &me, ":%s MODE %s -%c %s!%s@%s",
				 me.name, chptr->chname,
				 mask_index_char[type], mask->nick,
				 *mask->user ? mask->user : "*",
				 *mask->host ? mask->host : "*");
	  sendto_match_servs(chptr, &me, ":%s MODE %s -%c %s!%s@%s", me.name,
			     chptr->chname, mask_index_char[type],
			     mask->nick, *mask->user ? mask->user : "*",
			     *mask->host ? mask->host : "*");
	  *prev = cmlink->next;
	  free_link(cmlink);
	  cmlink = *prev;
	  continue;
	}

      /* OK, we got this far, this entry is valid. Is it what we want? */
      if (exact)
	{
	  if ((irccmp(mask->nick, search_mask->nick) == 0) &&
	      (irccmp(mask->user, search_mask->user) == 0) &&
	      (irccmp(mask->host, search_mask->host) == 0))
	    return cmlink;
	}
      else
	{
	  /* Check these in the order they are likely to be used */
	  if (match(mask->host, search_mask->host) &&
	      match(mask->nick, search_mask->nick) &&
	      match(mask->user, search_mask->user))
	    return cmlink;
	}
    }

  return NULL;
}

static void
add_channel_mask(struct Channel *chptr, const struct channel_mask *mask,
		 u_int32_t type)
{
  struct SLink *mask_entry;

  /* Don't add a mask exactly matching an existing one */
  if ((mask_entry = find_channel_mask(mask, chptr, type, 1, 0)))
    {
      /* But do update it's timeout/channel fields */
      mask_entry->value.cmptr->timeout = mask->timeout;
      strncpy_irc(mask_entry->value.cmptr->channel, mask->channel,
		  CHANNELLEN);
      return;
    }

  mask_entry = make_link();
  memset(mask_entry, 0, sizeof(struct SLink));
  mask_entry->value.cmptr =
    (struct channel_mask *) MyMalloc(sizeof(struct channel_mask));
  memcpy(mask_entry->value.cmptr, mask, sizeof(struct channel_mask));
  mask_entry->next = chptr->mode_mask_list[type];
  chptr->mode_mask_list[type] = mask_entry;
}

void
delete_channel_mask(struct Channel *chptr, struct SLink *cmlink,
		    u_int32_t type)
{
  MyFree(cmlink->value.cmptr);
  if (cmlink->next)
    {
      struct SLink *temp = cmlink->next;
      cmlink->value.cmptr = temp->value.cmptr;
      cmlink->next = temp->next;
      free_link(temp);
    }
  else
    cmlink->value.cmptr = NULL;
}

static void
list_channel_masks(struct Client *sptr, const struct Channel *chptr,
		   u_int32_t type)
{
  struct SLink *cmlink;
  for (cmlink = chptr->mode_mask_list[type]; cmlink; cmlink = cmlink->next)
    {
      struct channel_mask *mask = cmlink->value.cmptr;
      if (mask)
	sendto_one(sptr, form_str(mask_reply[type]),
		   me.name, sptr->name, chptr->chname,
		   mask->nick, mask->user, mask->host,
		   mask->placed_by, mask->placed_at,
		   mask->timeout, mask->reason);
    }
  sendto_one(sptr, form_str(mask_end_reply[type]), me.name, sptr->name,
	     chptr->chname);
}

/*
 *  Fixes a string so that the first white space found becomes an end of
 * string marker (`\0`).  returns the 'fixed' string or "*" if the string
 * was NULL length or a NULL pointer.
 */
static char *
check_string(char *s)
{
  static char star[2] = "*";
  char *str = s;

  if (BadPtr(s))
    return star;

  for (; *s; ++s)
    {
      if (IsSpace(*s))
	{
	  *s = '\0';
	  break;
	}
    }
  return str;
}

/* Here's an absurd parameter list for you. This is called by
 * m_sjoin() as well, so it's broken out of set_channel_mode()
 */

int
parse_mode_char(struct Client *sptr,
		struct Channel *chptr,
		char c,
		char plus, int whatt,
		char **mbufw, char **mbuf2w,
		char **pbufw, char **pbuf2w,
		int *len, int *parc, char ***parv)
{
  /* Grab these now to make it all easier to read */
  int cmode = cmode_index[(unsigned char) c];
  u_int32_t cm_type = cmodes[cmode].type;
  int position = cmodes[cmode].position;
  char *arg;
  register int arglen;
  int recalc = 0;

  if (whatt == MODE_QUERY)
    {
      if (cm_type == CMTYPE_MASK)
	list_channel_masks(sptr, chptr, cmodes[cmode].value);
      /* Otherwise there's no sensible way to query this */
      return recalc;
    }

  if (cmodes[cmode].umode_required
      && !IsServer(sptr) && !HasUmode(sptr, cmodes[cmode].umode_required))
    return recalc;

  /* Simple case first. Channel flags */
  if (cm_type == CMTYPE_CHANFLAG)
    {
      /* Abort if we don't have space to store the change */
      if (*len + 2 >= MODEBUFLEN)
	return recalc;

      if (whatt == MODE_ADD)
	{
	  chptr->mode_flags |= cmodes[cmode].value;
	  *(*mbufw)++ = '+';
	}
      else
	{
	  chptr->mode_flags &= ~cmodes[cmode].value;
	  *(*mbufw)++ = '-';
	}
      *(*mbufw)++ = c;
      *len += 2;

      /* Special cases */
      switch (c)
	{
	case 'j':
	  sendto_local_ops_flag(UMODE_SERVNOTICE, "%s %s channel %s",
				sptr->name,
				(whatt == MODE_ADD) ? "juping" : "unjuping",
				chptr->chname);
	  break;
	default:
	  break;
	}
      return recalc;
    }

  /* We may have a parameter. Here are the cases: */
  /* intval and strval only when adding
   * userflag and mask always
   */
  if ((((cm_type == CMTYPE_INTVAL) || (cm_type == CMTYPE_STRVAL))
       && whatt == MODE_ADD) || (cm_type == CMTYPE_USERFLAG)
      || (cm_type == CMTYPE_MASK))
    {
      if ((*parc)-- <= 0)
	{
	  if (MyClient(sptr))
	    {
	      static char buf[] = "MODE + ";
	      buf[5] = plus;
	      buf[6] = c;
	      sendto_one(sptr, form_str(ERR_NEEDMOREPARAMS),
			 me.name, sptr->name, buf);
	    }
	  return recalc;
	}

      arg = check_string(*(*parv)++);
    }

  /* Slightly trickier. Integer value flags */
  if (cm_type == CMTYPE_INTVAL)
    {
      char numeric[20];		/* Big enough for 2^64, for platforms where int is 64-bit */
      int n;

      if (whatt == MODE_ADD)
	{
	  n = atoi(arg);
	  if (n <= 0)
	    return recalc;

	  /* Reproduce the number to normalise it (strip leading 0s and suchlike) */
	  ircsprintf(numeric, "%d", n);
	  arg = numeric;
	  arglen = strlen(arg);

	  if (*len + 2 + arglen + 1 >= MODEBUFLEN)
	    return recalc;

	  if (position == 0)
	    {
	      strcpy((*pbufw), arg);
	      (*pbufw) += arglen;
	      *(*pbufw)++ = ' ';
	    }
	  else
	    {
	      strcpy((*pbuf2w), arg);
	      (*pbuf2w) += arglen;
	      *(*pbuf2w)++ = ' ';
	    }
	  *len += arglen + 1;
	}
      else
	{
	  if (*len + 2 >= MODEBUFLEN)
	    return recalc;
	  n = 0;
	}

      /* 0 is unset, anything else is set */
      chptr->mode_intval[cmodes[cmode].value] = n;

      if (position == 0)
	{
	  *(*mbufw)++ = plus;
	  *(*mbufw)++ = c;
	}
      else
	{
	  *(*mbuf2w)++ = plus;
	  *(*mbuf2w)++ = c;
	}
      *len += 2;

      switch (c)
	{
	default:
	  break;
	}
      return recalc;
    }
  /* ... and the companion string value types */
  if (cm_type == CMTYPE_STRVAL)
    {
      int index = cmodes[cmode].value;

      if (whatt == MODE_ADD)
	{
	  if (strlen(arg) > MODESTRLEN)
	    arg[MODESTRLEN] = '\0';
	  arglen = strlen(arg);

	  if (*len + 2 + arglen + 1 >= MODEBUFLEN)
	    return recalc;

	  if (chptr->mode_strval[index])
	    MyFree(chptr->mode_strval[index]);
	  DupString(chptr->mode_strval[index], arg);

	  if (position == 0)
	    {
	      strcpy((*pbufw), arg);
	      (*pbufw) += arglen;
	      *(*pbufw)++ = ' ';
	    }
	  else
	    {
	      strcpy((*pbuf2w), arg);
	      (*pbuf2w) += arglen;
	      *(*pbuf2w)++ = ' ';
	    }
	  *len += arglen + 1;
	}
      else
	{
	  if (*len + 2 >= MODEBUFLEN)
	    return recalc;

	  MyFree(chptr->mode_strval[index]);
	  chptr->mode_strval[index] = NULL;
	}

      if (position == 0)
	{
	  *(*mbufw)++ = plus;
	  *(*mbufw)++ = c;
	}
      else
	{
	  *(*mbuf2w)++ = plus;
	  *(*mbuf2w)++ = c;
	}
      *len += 2;

      switch (c)
	{
	default:
	  break;
	}
      return recalc;
    }
  /* More complex, user flags */
  if (cm_type == CMTYPE_USERFLAG)
    {
      struct Client *acptr;	/* Target of this mode change */
      acptr = find_chasing(sptr, arg, NULL);
      /* I don't really think this can happen, but... */
      if (!acptr->user)
	{
	  sendto_ops_flag(UMODE_DEBUG,
			  "WTF: find_chasing() gave me something without a User struct (%s)",
			  acptr->name);
	  return recalc;
	}

      if (!IsMember(acptr, chptr))
	{
	  if (MyClient(sptr))
	    sendto_one(sptr, form_str(ERR_USERNOTINCHANNEL), me.name,
		       sptr->name, arg, chptr->chname);
	  return recalc;
	}

      switch (c)
	{
	case 'o':
	  if ((whatt == MODE_DEL) && (acptr != sptr)
	      && HasUmode(acptr, UMODE_IMMUNE) && !HasUmode(sptr, UMODE_GOD))
	    {
	      /* This user is immune. Go away. */
	      sendto_one(sptr, form_str(ERR_USERISIMMUNE),
			 me.name, sptr->name, acptr->name, chptr->chname);
	      return recalc;
	    }
	  if (acptr == sptr)
	    recalc = 1;		/* This user changed chanop status; recalc if in set_channel_mode() */
	  break;
	default:
	  break;
	}

      arg = acptr->name;
      arglen = strlen(arg);
      if (*len + 2 + arglen + 1 >= MODEBUFLEN)
	return recalc;

      change_chan_flag(chptr, acptr, cmodes[cmode].value, whatt);

      if (position == 0)
	{
	  *(*mbufw)++ = plus;
	  *(*mbufw)++ = c;
	  strcpy((*pbufw), arg);
	  (*pbufw) += arglen;
	  *(*pbufw)++ = ' ';
	}
      else
	{
	  *(*mbuf2w)++ = plus;
	  *(*mbuf2w)++ = c;
	  strcpy((*pbuf2w), arg);
	  (*pbuf2w) += arglen;
	  *(*pbuf2w)++ = ' ';
	}
      *len += 2 + arglen + 1;

      return recalc;
    }
  /* Lastly, channel masks */
  if (cm_type == CMTYPE_MASK)
    {
      struct channel_mask cmask;
      if (parse_channel_mask(arg, &cmask, parc, parv) != 0)
	return recalc;

      cmask.placed_at = CurrentTime;
      strncpy_irc(cmask.placed_by, sptr->name, NICKLEN);

      if (whatt == MODE_ADD)
	add_channel_mask(chptr, &cmask, cmodes[cmode].value);
      else
	{
	  struct SLink *cmlink;
	  if ((cmlink =
	       find_channel_mask(&cmask, chptr, cmodes[cmode].value, 1,
				 MyClient(sptr))))
	    delete_channel_mask(chptr, cmlink, cmodes[cmode].value);
	  else
	    return recalc;
	}

      switch (c)
	{
	default:
	  break;
	}

      arg = build_channel_mask(&cmask);
      arglen = strlen(arg);

      if (position == 0)
	{
	  *(*mbufw)++ = plus;
	  *(*mbufw)++ = c;
	  strcpy((*pbufw), arg);
	  (*pbufw) += arglen;
	  *(*pbufw)++ = ' ';
	}
      else
	{
	  *(*mbuf2w)++ = plus;
	  *(*mbuf2w)++ = c;
	  strcpy((*pbuf2w), arg);
	  (*pbuf2w) += arglen;
	  *(*pbuf2w)++ = ' ';
	}

      return recalc;
    }
  return recalc;
}

void
set_channel_mode(struct Client *cptr,
		 struct Client *sptr,
		 struct Channel *chptr, int parc, char *parv[])
{
  char *current_parameter, c;
  char *mbufw = modebuf, *mbuf2w = modebuf2;
  char *pbufw = parabuf, *pbuf2w = parabuf2;
  int len = 0;

  int whatt = MODE_QUERY;
  char plus = '=';

  int user_mode = user_channel_mode(sptr, chptr);
  int is_chanop = IsServer(sptr) || (user_mode & CUFLAG_CHANOP);
  /* Precalculate and store these, I need to query them
   * several times
   *  -- asuffield
   */
  int is_member = IsMember(sptr, chptr);
  int is_god = IsServer(sptr) || HasUmode(sptr, UMODE_GOD);

  int can_change = (is_chanop || is_god) && !chptr->frozen_by;
  int can_query = is_member || is_god;

  /* Refuse all mode changes during a channel freeze */
  if (chptr->frozen_by && MyClient(sptr))
    {
      sendto_one(sptr, form_str(RPL_LOAD2HI), me.name, sptr->name);
      return;
    }

  /* This function is called with the first two parameters
   * (origin and channel names) already removed. parv[0] is therefore
   * the first mode change thing
   */
  current_parameter = parv[0];
  /* current_parameter always contains the next string which
   * is to be parsed as a series of mode letters. parv[0] is always
   * the _next_ parameter which has not yet been processed
   */
  parc--;
  parv++;

  for (;;)
    {
      if (BadPtr(current_parameter))
	{
	  /* We've run out of modes in this parameter. Do we have
	   * any more parameters we haven't processed yet?
	   */
	  if (parc-- > 0)
	    {
	      current_parameter = *parv++;
	      /* Reset whatt so that this:
	       *  +b foo!*@* bar*!*@*
	       * doesn't try to set cmodes +bar*!@
	       *
	       * Also means that +b foo b works as expected
	       */
	      whatt = MODE_QUERY;
	      continue;
	    }
	  else
	    /* No. We're done.
	     * NOTE: This is the only exit point from this for() loop
	     *       unless parsing is aborted due to an error
	     */
	    break;
	}
      /* OK, if we get to here, current_parameter contains
       * something we are supposed to parse as a mode
       * string
       */
      /* Grab the first character */
      c = *current_parameter++;

      switch (c)
	{
	case '+':
	  whatt = MODE_ADD;
	  plus = '+';
	  continue;

	case '-':
	  whatt = MODE_DEL;
	  plus = '-';
	  continue;

	case '=':
	  whatt = MODE_QUERY;
	  plus = '=';
	  continue;

	default:
	  if ((whatt == MODE_QUERY) && !can_query)
	    continue;
	  if (((whatt == MODE_ADD) || (whatt == MODE_DEL)) && !can_change)
	    continue;

	  if (parse_mode_char(sptr, chptr, c, plus, whatt,
			      &mbufw, &mbuf2w, &pbufw, &pbuf2w,
			      &len, &parc, &parv))
	    {
	      /* Recalculate the stuff about whether or not this
	       * user is a chanop
	       */
	      is_chanop = (whatt == MODE_ADD);
	      can_change = (is_chanop || is_god) && !chptr->frozen_by;
	      can_query = is_member || is_god;
	    }

	  break;
	}
    }

  /* OK, combine the buffers */
  /* Do this now for modebuf so that the signs get collapsed together.
   * parabuf can be handled in the sendto in a bit.
   */
  *mbuf2w = *pbufw = *pbuf2w = '\0';
  strncpy_irc(mbufw, modebuf2, MODEBUFLEN - strlen(modebuf));
  collapse_signs(modebuf);

  if (*modebuf)			/* Actually, this could be nil if all modes were invalid */
    {
      sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s %s",
			     IsServer(sptr) ? me.name : sptr->name,
			     chptr->chname, modebuf, parabuf, parabuf2);
      sendto_serv_butone(cptr, ":%s MODE %s %s %s %s",
			 sptr->name, chptr->chname,
			 modebuf, parabuf, parabuf2);
    }
  return;
}

void
channel_modes(const struct Client *cptr, char *mbuf, char *pbuf,
	      const struct Channel *chptr, int all)
{
  int position, i;
  char *modebuf = mbuf;

  *mbuf++ = '+';

  /* This iterates twice. Rewrite it sometime. */
  for (position = 0; position < 2; position++)
    for (i = 0; cmodes[i].value; i++)
      if (cmodes[i].position == position)
	switch (cmodes[i].type)
	  {
	  case CMTYPE_CHANFLAG:
	    if (chptr->mode_flags & cmodes[i].value)
	      {
		*mbuf++ = '+';
		*mbuf++ = cmodes[i].letter;
	      }
	    else if (all)
	      {
		*mbuf++ = '-';
		*mbuf++ = cmodes[i].letter;
	      }
	    break;
	  case CMTYPE_INTVAL:
	    if (chptr->mode_intval[cmodes[i].value])
	      {
		*mbuf++ = '+';
		*mbuf++ = cmodes[i].letter;
		pbuf +=
		  ircsprintf(pbuf, "%d ",
			     chptr->mode_intval[cmodes[i].value]);
	      }
	    else if (all)
	      {
		*mbuf++ = '-';
		*mbuf++ = cmodes[i].letter;
	      }
	    break;
	  case CMTYPE_STRVAL:
	    if (chptr->mode_strval[cmodes[i].value])
	      {
		*mbuf++ = '+';
		*mbuf++ = cmodes[i].letter;
		pbuf +=
		  ircsprintf(pbuf, "%s ",
			     chptr->mode_strval[cmodes[i].value]);
	      }
	    else if (all)
	      {
		*mbuf++ = '-';
		*mbuf++ = cmodes[i].letter;
	      }
	    break;
	  case CMTYPE_USERFLAG:
	  case CMTYPE_MASK:
	    /* Don't show these ones */
	    break;
	  default:
	    sendto_ops_flag(UMODE_DEBUG,
			    "WTF: Unknown cm_type %d in cmodes[%d]",
			    cmodes[i].type, i);
	  }
  *mbuf = *pbuf = '\0';

  collapse_signs(modebuf);
}

int
user_channel_mode(struct Client *cptr, struct Channel *chptr)
{
  Link *lp;

  if (chptr)
    if ((lp = find_user_link(chptr->members, cptr)))
      return (lp->flags);

  return 0;
}

void
change_chan_flag(struct Channel *chptr, struct Client *cptr, int flag,
		 int whatt)
{
  Link *tmp;

  if ((tmp = find_user_link(chptr->members, cptr)))
    {
      if (whatt == MODE_ADD)
	tmp->flags |= flag;
      else
	tmp->flags &= ~flag;
    }
}

struct SLink *
match_channel_mask(const struct channel_mask *mask, struct Channel *chptr,
		   char mode)
{
  int cmode = cmode_index[(unsigned char) mode];
  return find_channel_mask(mask, chptr, cmodes[cmode].value, 0, 1);
}

/* This just frees any allocated memory in the mode structures */
void
free_channel_modes(struct Channel *chptr)
{
  struct SLink *next, *target;
  int i;

  for (i = 0; cmodes[i].value; i++)
    switch (cmodes[i].type)
      {
      case CMTYPE_CHANFLAG:
      case CMTYPE_INTVAL:
      case CMTYPE_USERFLAG:
	break;
	break;
      case CMTYPE_STRVAL:
	MyFree(chptr->mode_strval[cmodes[i].value]);
	break;
      case CMTYPE_MASK:
	for (next = chptr->mode_mask_list[cmodes[i].value]; (target = next);)
	  {
	    next = target->next;
	    MyFree(target->value.cmptr);
	    free_link(target);
	  }
	break;
      default:
	sendto_ops_flag(UMODE_DEBUG, "WTF: Unknown cm_type %d in cmodes[%d]",
			cmodes[i].type, i);
      }
}

int
get_mode_intval(struct Channel *chptr, char mode)
{
  int cmode = cmode_index[(unsigned char) mode];
  assert(cmodes[cmode].type == CMTYPE_INTVAL);
  return chptr->mode_intval[cmodes[cmode].value];
}

char *
get_mode_strval(struct Channel *chptr, char mode)
{
  int cmode = cmode_index[(unsigned char) mode];
  assert(cmodes[cmode].type == CMTYPE_STRVAL);
  return chptr->mode_strval[cmodes[cmode].value];
}
